mod error;

use super::Frequency as RequestedFrequencyType;
use rubato::{FftFixedIn, Resampler};
use symphonia::core::{
    audio::{Layout, SampleBuffer},
    codecs::{Decoder, DecoderOptions, CODEC_TYPE_NULL},
    errors::Error as SymError,
    formats::{FormatOptions, FormatReader},
    io::MediaSourceStream,
    meta::MetadataOptions,
    probe::Hint, sample
};
use tool_helper::{Error, Input};

#[derive(Copy, Clone, PartialEq)]
pub enum Orality {
    Stereo,
    Mono,
}

impl Orality {
    fn as_channels(&self) -> usize {
        match self {
            Orality::Mono   => 1,
            Orality::Stereo => 2,
        }
    }
}

pub struct AudioSamples<T> {
    samples:   Vec::<T>,
    orality:   Orality,
    frequency: u32,
}

impl<T> AudioSamples<T> {
    pub fn new(samples: Vec<T>, channels: usize, frequency: u32) -> Result<AudioSamples<T>, Error> {
        let orality = match channels {
            0 => return Err(Error::from_str("Input file has no audio channels")),
            1 => Orality::Mono,
            2 => Orality::Stereo,
            _ => return Err(Error::from_str("Only Mono and Stereo input are supported")),
        };
        Ok(AudioSamples{samples, orality, frequency})
    }

    // TODO: Reconsider passing zero_maker
    pub fn to_planar(self, chunk_size: usize, zero_maker: fn()->T) -> Vec<Vec::<T>> {
        let (_, chunk_remainder) = self.calculate_chunks(chunk_size);
        let num_channel          = self.orality.as_channels();
        let mut result           = Vec::new();

        for _ in 0..num_channel {
            result.push(Vec::<T>::new());
        }
        
        for (idx, sample) in self.samples.into_iter().enumerate() {
            result[idx%num_channel].push(sample);
        }

        for channel in &mut result {
            println!("({} - {} = {})", chunk_size, chunk_remainder, chunk_size - chunk_remainder);
            for _ in 0..(chunk_size - chunk_remainder) {
                channel.push(zero_maker());
            }
        }

        result
    }

    pub fn calculate_chunks(&self, chunk_size: usize) -> (usize, usize) {
        let sample_len = self.samples.len()/self.orality.as_channels();
        let div_value  = chunk_size;

        (sample_len/div_value, sample_len%div_value)
    }

    pub fn samples(&self) -> &Vec::<T> {
        &self.samples
    }

    pub fn orality(&self) -> Orality {
        self.orality.clone()
    }
}

struct InternalAudioSamples {
    planar_samples: Vec<Vec<f32>>,
    frequency:      u32,
}

impl InternalAudioSamples {
    pub fn new(planar_samples: Vec<Vec<f32>>, frequency: u32) -> InternalAudioSamples {
        InternalAudioSamples{planar_samples, frequency}
    }

    pub fn calculate_chunks(&self, chunk_size: usize) -> (usize, usize) {
        if self.planar_samples.len() > 0 {
            let sample_len = self.planar_samples[0].len();
            (sample_len/chunk_size, sample_len%chunk_size)
        }

        else {
            (0, 0)
        }
    }

    pub fn channels(&self) -> usize {
        self.planar_samples.len()
    }

    pub fn planar_slices(&self) -> Vec<&[f32]> {
        let mut planar_slices = Vec::new();

        for channel in &self.planar_samples {
            planar_slices.push(channel.as_slice());
        }
        planar_slices
    }
}

pub type PreparedAudioSamples = AudioSamples<i16>;

pub fn load_as_i16_audio(input: Input, freq_type: RequestedFrequencyType) -> Result<PreparedAudioSamples, Error> {
    let raw_audio = load_raw_audio(input)?;
    let raw_audio = resample(raw_audio, 37_800)?;

    test_write_wav(raw_audio)?;
    Err(Error::not_implemented("Resampling not implemented"))
}

fn test_write_wav(audio_samples: InternalAudioSamples) -> Result<(), Error> {
    /*let spec = hound::WavSpec {
        channels: 2,
        sample_rate: 37_800,
        bits_per_sample: 32,
        sample_format: hound::SampleFormat::Float,
    };

    let mut file = hound::WavWriter::create("planschi.wav", spec).unwrap();
    for sample in audio_samples.samples {
        file.write_sample(sample)?;
    }
    file.finalize()?;*/
    Ok(())
}

fn load_raw_audio(input: Input) -> Result<InternalAudioSamples, Error> {
    let media_stream = MediaSourceStream::new(Box::new(load_to_ram(input)?), Default::default());
    let format       = symphonia::default::get_probe().format(&Hint::new(), media_stream, &FormatOptions::default(), &MetadataOptions::default()).map_err(error::probe)?.format;
    let track        = format.tracks().iter().find(|t| t.codec_params.codec != CODEC_TYPE_NULL).ok_or_else(error::find_track)?;
    
    // Create a decoder for the track.
    let decoder  = symphonia::default::get_codecs().make(&track.codec_params, &DecoderOptions::default()).map_err(error::decoder)?;
    let track_id = track.id;

    decode(format, decoder, track_id)
}

fn decode(mut format: Box<dyn FormatReader>, mut decoder: Box<dyn Decoder>, track_id: u32) -> Result<InternalAudioSamples, Error> {
    let mut samples       = Vec::new();
    let mut channel_count = 0;
    let mut frequency     = 0;
    let mut read_buffer   = None;

    loop {
        // Get the next packet from the media format.
        let packet = match format.next_packet() {
            Ok(packet) => packet,
            Err(err) => {
                if let SymError::IoError(io_err) = &err {
                    if io_err.kind() == std::io::ErrorKind::UnexpectedEof {
                        return Ok(InternalAudioSamples::new(samples, frequency));
                    }                    
                }
                return Err(error::next_packet(err));
            }
        };

        // Consume any new metadata that has been read since the last packet.
        format.metadata().skip_to_latest();

        // If the packet does not belong to the selected track, skip over it.
        if packet.track_id() != track_id {
            continue;
        }

        // Decode the packet into audio samples.
        let packet = decoder.decode(&packet).map_err(error::decode)?;
        if read_buffer.is_none() {
            let duration = packet.capacity() as u64;
            let specs    = packet.spec();

            channel_count = specs.channels.count();
            frequency     = specs.rate;
            read_buffer   = Some(SampleBuffer::<f32>::new(duration, packet.spec().clone()));
            
            for _ in 0..channel_count {
                samples.push(Vec::new());
            }
        }

        if let Some(read_buffer) = &mut read_buffer {
            read_buffer.copy_planar_ref(packet);
            let     cur_samples = read_buffer.samples();
            let mut cur_samples = cur_samples.chunks(cur_samples.len()/channel_count);

            for dst_sample in &mut samples {
                dst_sample.extend(cur_samples.next().ok_or_else(|| Error::from_str("Not enough channels in input as expected"))?);
            }
        }        
    }
}

fn resample(input: InternalAudioSamples, target_frequency: u32) -> Result<InternalAudioSamples, Error> {
    const HIGH_QUALITY_CHUNKS:usize = (1024*10)*100;
    const SUB_CHUNKS:usize          = 20;

    let     (chunk_size, sub_chunks)       = (HIGH_QUALITY_CHUNKS, SUB_CHUNKS);
    let     (chunk_count, chunk_remainder) = input.calculate_chunks(chunk_size);
    let mut planar_input                   = input.planar_slices();
    let mut resampler                      = FftFixedIn::<f32>::new(input.frequency as usize, target_frequency as usize, chunk_size, sub_chunks, input.channels()).map_err(error::resampler_construction)?;
    let mut planar_output                  = {
        let mut planar_output = Vec::new();

        for _ in 0..planar_input.len() {
            planar_output.push(Vec::<f32>::new());
        }
        planar_output
    };

    for _chunk in 0..chunk_count {
        let new_samples = resampler.process(&planar_input, None).map_err(error::resample)?;
        for (channel, slice) in planar_input.iter_mut().enumerate() {
            *slice = &slice[chunk_size..];
            planar_output[channel].extend(new_samples[channel].iter());
        }
    }

    let mut process_partial = |input_option: Option<&[&[f32]]>| -> Result<(), Error> {
        let new_samples = resampler.process_partial(input_option, None).map_err(error::resample)?;
        for (channel, channel_samples) in new_samples.into_iter().enumerate() {
            planar_output[channel].extend(channel_samples.iter());
        }

        Ok(())
    };

    if chunk_remainder > 0 {
        process_partial(Some(&planar_input))?;
    }
    process_partial(None)?;

    ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    let spec = hound::WavSpec {
        channels: 1,
        sample_rate: 37_800,
        bits_per_sample: 32,
        sample_format: hound::SampleFormat::Float,
    };

    let mut file = hound::WavWriter::create("planschi.wav", spec).unwrap();
    for sample in &planar_output[0] {
        file.write_sample(*sample)?;
    }
    file.finalize()?;


    Err(Error::not_implemented("resample"))
}

fn load_to_ram(mut input: Input) -> Result<std::io::Cursor<Vec<u8>>, Error> {
    let mut buffer = Vec::default();
    
    input.read_to_end(&mut buffer)?;
    Ok(std::io::Cursor::new(buffer))
}